import requests
import json
import os
from pathlib import Path
from urllib.parse import urljoin
from getpass import getpass

LOCALHOST = "http://localhost:8000"
DEFAULT_CREDENTIALS = os.path.expanduser('~/.mwc')

class MakingWithCodeAPI:
    def __init__(self, host=LOCALHOST, credentials_file=DEFAULT_CREDENTIALS):
        self.host = host
        self.credentials_file = Path(credentials_file)

    def authenticate(self):
        "Sets self.token, either from cached credentials or by prompting for username/password"
        if self.credentials_file.exists():
            credentials = json.loads(self.credentials_file.read_text())
            if credentials.get('token'):
                self.token = credentials['token']
                return
        else:
            credentials = {}
        username = input("username: ")
        password = getpass("password: ")
        credentials['token'] = self.get_token(username, password)
        self.token = credentials['token']
        self.credentials_file.write_text(json.dumps(credentials))

    def get_token(self, username, password):
        url = urljoin(self.host, "api/token/")
        result = requests.post(url, data={"username": username, "password": password})
        return self.handle_response(result)['token']

    def get(self, urlpath):
        "Makes a GET request with appropriate headers and params, returning JSON"
        self.authenticate()
        url = urljoin(self.host, urlpath + '/')
        headers = {"Authorization": "Token {}".format(self.token)}
        params = {"format": "json"}
        response = requests.get(url, headers=headers, params=params)
        return self.handle_response(response)

    def post(self, urlpath, payload=None):
        "Makes a POST request with given payload, returning JSON"
        self.authenticate()
        url = urljoin(self.host, urlpath + '/')
        response = requests.post(url, data=payload)
        return self.handle_response(response)

    def handle_response(self, response):
        if response.ok:
            return response.json()
        elif response.status_code == 500:
            raise self.RequestFailed("Error 500")
        else:
            raise self.RequestFailed(response.json())

    class RequestFailed(Exception):
        pass








from .github_backend import GithubBackend
from .github_org_backend import GithubOrgBackend

def get_backend(name):
    return {
        'github': GithubBackend,
        'github_org': GithubOrgBackend,
    }[name]



from textwrap import fill
import click

FW = 80

def address(message, preformatted=False, list_format=False):
    """Addresses the user (e.g. a greeting or transition to the next step).
    This may be verbose, so text is wrapped.
    """
    if preformatted:
        if list_format: 
            raise ValueError("preformatted and list_format are incompatible options")
        fmsg = message
    elif list_format:
        fmsg = fill(message, width=FW, initial_indent='- ', subsequent_indent='  ')
    else:
        fmsg = fill(message, width=FW) + '\n'
    return click.style(fmsg, fg='cyan')

def question(message):
    """Addresses the user qith a question (e.g. a prompt)
    """
    return click.style(fill(message, width=FW) + '\n', fg="cyan")

def debug(message, preformatted=False):
    """Shows the user debug information"""
    if preformatted:
        return click.style(message, fg="gray")
    else:
        return click.style(fill(message, width=FW), fg="gray")

def info(message, preformatted=False):
    """Shows the user information. 
    Don't use this for boilerplate; the user should have requested the information 
    or may need to check it.
    """
    if preformatted:
        return click.style(message, fg="blue")
    else:
        return click.style(fill(message, width=FW), fg="blue")

def warn(message):
    """Warns the user."""
    return click.style(fill(message, width=FW), fg="yellow")

def confirm(message):
    """Asks the user to confirm a potentially dangerous action.
    """
    return warn(message)

def error(message):
    return click.style(fill(message, width=FW), fg="red")

def success(message):
    return click.style(fill(message, width=FW), fg="green")



from .base_backend import GitBackend
from subprocess import run
from pathlib import Path
import click
import json
import os
from making_with_code_cli.helpers import cd
from making_with_code_cli.styles import (
    confirm,
)

COMMIT_TEMPLATE = ".commit_template"

class GithubOrgBackend(GitBackend):
    """A Github backend which uses github organizations.  
    Students own their own repos but create them from within the organization.
    """

    @classmethod
    def extend_settings(self, settings):
        "Asks for the user's github username and email address"
        def choose_github_username(github_username):
            msg = "What is your GitHub username? "
            return click.prompt(msg, default=github_username)
        def choose_github_org(github_org):
            msg = "What is the name of the course's GitHub organization? "
            return click.prompt(msg, default=github_org)
        def choose_github_email(github_email):
            msg = "What is the email address associated with your GitHub account? "
            return click.prompt(msg, default=github_email)
        def choose_git_name(git_name):
            msg = "What name do you want to use in your git commits? "
            return click.prompt(msg, default=git_name)
        if settings.get('github_username') or click.confirm("Do you have a GitHub account?"):
            settings['github_username'] = choose_github_username(settings.get("github_username"))
            settings['github_org'] = choose_github_org(settings.get("github_org"))
            settings['github_email'] = choose_github_email(settings.get("github_email"))
            settings['git_name'] = choose_git_name(settings.get("git_name"))
        return settings

    def init_clone(self, module, modpath):
        url = module["repo_url"]
        cmd = f'gh repo clone "{url}" "{modpath.name}"'
        with cd(modpath.parent):
            run(cmd, shell=True, check=True)
        if (modpath / COMMIT_TEMPLATE).exists():
            with cd(modpath):
                run(f"git config commit.template {COMMIT_TEMPLATE}")

    def init_create_from_template(self, module, modpath):
        """Creates the named repo from a template, or clones an existing repo with. 
        """
        repo_name = self.get_repo_name_from_template_repo_url(module["repo_url"])
        url = module["repo_url"]

        if self.user_has_repo(repo_name):
            cmd = f'gh repo clone "{url}" "{modpath.name}"'
        else:
            if modpath.exists():
                self.relocate_existing_directory(modpath)
            usr = self.settings['username']
            org = self.settings['github_org']
            cmd = f'gh repo create "{org}/{repo_name}" --clone --private --template "{url}"'
        with cd(modpath.parent):
            run(cmd, shell=True, check=True)
            run(f"mv {repo_name} {modpath.name}", shell=True, check=True) 
        if (modpath / COMMIT_TEMPLATE).exists():
            with cd(modpath):
                run(f"git config commit.template {COMMIT_TEMPLATE}")

    def get_repo_name_from_template_repo_url(self, url):
        """Parses the template repo URL and returns the name of a repo to create.
        Expects a GitHub url like "https://github.com/cproctor/mwc-pedprog-unit00-project-drawing.git"
        """
        parts = url.split('/')
        name, suffix = parts[-1][:-4], parts[-1][-4:]
        return name + '-' + self.settings['github_username']

    def relocate_existing_directory(self, path):
        """Moves an existing directory out of the way.
        """
        new_path = path.parent / path.name + '_old'
        while new_path.exists():
            new_path = new_path.parent / new_path.name + '_old'
        click.echo(confirm(f"Moving existing directory {path} to {new_path}."))
        os.rename(path, new_path)

    def user_has_repo(self, name):
        "Checks to see whether the user already has the named repo."
        org = self.settings['github_org']
        cmd = f"gh repo list {org} --json name --limit 10000"
        result = run(cmd, shell=True, capture_output=True, text=True).stdout
        repo_names = [obj['name'] for obj in json.loads(result)]
        return name in repo_names



import requests

def get_curriculum(settings):
    url = settings["mwc_site_url"] + "/index.json"
    response = requests.get(url)
    if response.ok:
        return response.json()
    else:
        raise IOError(f"Error getting curriculum from {url}.")



from pathlib import Path
from subprocess import run, CalledProcessError
import click
from making_with_code_cli.helpers import cd
from making_with_code_cli.cli_setup import WORK_DIR_PERMISSIONS
from making_with_code_cli.styles import (
    address,
    confirm,
    error,
)

class GitBackend:
    """Base class interface to backend git server.
    All Making With Code deployments are backed by a git server, but the nature of the
    server and the strategies for completing tasks vary by backend. 
    """
    INIT_ACTIONS = [
        'create_from_template',
        'clone',
        'mkdir'
    ]

    @classmethod
    def extend_settings(self, settings):
        "A hook for git backends to collect additional information from the user"
        return settings

    def __init__(self, settings):
        self.settings = settings

    def init_module(self, module, modpath):
        if not 'init_action' in module:
            raise ValueError(f"There is a problem with the website. Can't initialize module {module['slug']} " + 
                    "without an init action.")
        if not module['init_action'] in self.INIT_ACTIONS:
            raise ValueError(f"There is a problem with the website. Can't initialize module {module['slug']} " + 
                    f"with init action '{module['init_action']}'.")
        if module['init_action'] == "mkdir":
            self.init_mkdir(module, modpath)
        if module['init_action'] == "clone":
            self.init_clone(module, modpath)
        if module['init_action'] == "create_from_template":
            self.init_create_from_template(module, modpath)

    def init_mkdir(self, module, modpath):
        modpath.mkdir(mode=WORK_DIR_PERMISSIONS)

    def init_clone(self, module, modpath):
        raise NotImplemented()

    def init_create_from_template(self, module, modpath):
        raise NotImplemented()

    def update(self, module, modpath):
        if (modpath / ".git").is_dir():
            with cd(modpath):
                relpath = self.relative_path(modpath)
                try:
                    click.echo(address(f"Checking {relpath} for updates.", preformatted=True))
                    run("git pull", shell=True, check=True)
                    if Path("pyproject.toml").exists():
                        run("poetry update", shell=True, check=True)
                except CalledProcessError as e:
                    click.echo(error(f"There was a problem updating {relpath}. Ask a teacher."))
                    raise e

    def work_dir(self):
        return Path(self.settings['work_dir'])

    def relative_path(self, path):
        return path.relative_to(self.work_dir())

